@@ -40,6 +40,7 @@ import org.apache.catalina.core.StandardContext;
 import org.apache.catalina.ha.CatalinaCluster;
 import org.apache.catalina.ha.ClusterMessage;
 import org.apache.catalina.ha.tcp.ReplicationValve;
+import org.apache.catalina.session.ManagerBase;
 import org.apache.catalina.tribes.Member;
 import org.apache.catalina.tribes.io.ReplicationStream;
 import org.apache.catalina.util.LifecycleSupport;
@@ -62,7 +63,7 @@ import org.apache.catalina.ha.ClusterManager;
  * @author Craig R. McClanahan
  * @author Jean-Francois Arcand
  * @author Peter Rossbach
- * @version $Revision: 674750 $ $Date: 2008-07-08 11:45:42 +0200 (Tue, 08 Jul 2008) $
+ * @version $Id: DeltaManager.java 1152894 2011-08-01 19:10:01Z markt $
  */
 
 public class DeltaManager extends ClusterManagerBase{
@@ -112,7 +113,8 @@ public class DeltaManager extends ClusterManagerBase{
     private boolean expireSessionsOnShutdown = false;
     private boolean notifyListenersOnReplication = true;
     private boolean notifySessionListenersOnReplication = true;
-    private boolean stateTransfered = false ;
+    private boolean notifyContainerListenersOnReplication  = true;
+    private volatile boolean stateTransfered = false ;
     private int stateTransferTimeout = 60;
     private boolean sendAllSessions = true;
     private boolean sendClusterDomainOnly = true ;
@@ -133,22 +135,23 @@ public class DeltaManager extends ClusterManagerBase{
     private long sessionReplaceCounter = 0 ;
     long processingTime = 0;
     private long counterReceive_EVT_GET_ALL_SESSIONS = 0 ;
-    private long counterSend_EVT_ALL_SESSION_DATA = 0 ;
     private long counterReceive_EVT_ALL_SESSION_DATA = 0 ;
     private long counterReceive_EVT_SESSION_CREATED = 0 ;
     private long counterReceive_EVT_SESSION_EXPIRED = 0;
     private long counterReceive_EVT_SESSION_ACCESSED = 0 ;
     private long counterReceive_EVT_SESSION_DELTA = 0;
+    private int counterReceive_EVT_ALL_SESSION_TRANSFERCOMPLETE = 0 ;
+    private long counterReceive_EVT_CHANGE_SESSION_ID = 0 ;
     private long counterSend_EVT_GET_ALL_SESSIONS = 0 ;
+    private long counterSend_EVT_ALL_SESSION_DATA = 0 ;
     private long counterSend_EVT_SESSION_CREATED = 0;
     private long counterSend_EVT_SESSION_DELTA = 0 ;
     private long counterSend_EVT_SESSION_ACCESSED = 0;
     private long counterSend_EVT_SESSION_EXPIRED = 0;
     private int counterSend_EVT_ALL_SESSION_TRANSFERCOMPLETE = 0 ;
-    private int counterReceive_EVT_ALL_SESSION_TRANSFERCOMPLETE = 0 ;
+    private long counterSend_EVT_CHANGE_SESSION_ID = 0;
     private int counterNoStateTransfered = 0 ;
 
-
     // ------------------------------------------------------------- Constructor
     public DeltaManager() {
         super();
@@ -224,7 +227,14 @@ public class DeltaManager extends ClusterManagerBase{
     public int getCounterSend_EVT_ALL_SESSION_TRANSFERCOMPLETE() {
         return counterSend_EVT_ALL_SESSION_TRANSFERCOMPLETE;
     }
- 
+
+    /**
+     * @return Returns the counterSend_EVT_CHANGE_SESSION_ID.
+     */
+    public long getCounterSend_EVT_CHANGE_SESSION_ID() {
+        return counterSend_EVT_CHANGE_SESSION_ID;
+    }
+
     /**
      * @return Returns the counterReceive_EVT_ALL_SESSION_DATA.
      */
@@ -274,6 +284,13 @@ public class DeltaManager extends ClusterManagerBase{
     public int getCounterReceive_EVT_ALL_SESSION_TRANSFERCOMPLETE() {
         return counterReceive_EVT_ALL_SESSION_TRANSFERCOMPLETE;
     }
+
+    /**
+     * @return Returns the counterReceive_EVT_CHANGE_SESSION_ID.
+     */
+    public long getCounterReceive_EVT_CHANGE_SESSION_ID() {
+        return counterReceive_EVT_CHANGE_SESSION_ID;
+    }
     
     /**
      * @return Returns the processingTime.
@@ -463,6 +480,14 @@ public class DeltaManager extends ClusterManagerBase{
         this.notifyListenersOnReplication = notifyListenersOnReplication;
     }
 
+    public boolean isNotifyContainerListenersOnReplication() {
+        return notifyContainerListenersOnReplication;
+    }
+
+    public void setNotifyContainerListenersOnReplication(
+            boolean notifyContainerListenersOnReplication) {
+        this.notifyContainerListenersOnReplication = notifyContainerListenersOnReplication;
+    }
     
     /**
      * @return Returns the defaultMode.
@@ -604,6 +629,66 @@ public class DeltaManager extends ClusterManagerBase{
         return new DeltaSession(this);
     }
 
+    /**
+     * Change the session ID of the current session to a new randomly generated
+     * session ID.
+     * 
+     * @param session   The session to change the session ID for
+     */
+    @Override
+    public void changeSessionId(Session session) {
+        changeSessionId(session, true);
+    }
+
+    public void changeSessionId(Session session, boolean notify) {
+        // original sessionID
+        String orgSessionID = session.getId();
+        super.changeSessionId(session);
+        if (notify) {
+            // changed sessionID
+            String newSessionID = session.getId();
+            try {
+                // serialize sessionID
+                byte[] data = serializeSessionId(newSessionID);
+                // notify change sessionID
+                SessionMessage msg = new SessionMessageImpl(getName(),
+                        SessionMessage.EVT_CHANGE_SESSION_ID, data,
+                        orgSessionID, orgSessionID + "-"
+                                + System.currentTimeMillis());
+                msg.setTimestamp(System.currentTimeMillis());
+                counterSend_EVT_CHANGE_SESSION_ID++;
+                send(msg);
+            } catch (IOException e) {
+                log.error(sm.getString("deltaManager.unableSerializeSessionID",
+                        newSessionID), e);
+            }
+        }
+    }
+
+    /**
+     * serialize sessionID
+     * @throws IOException if an input/output error occurs
+     */
+    protected byte[] serializeSessionId(String sessionId) throws IOException {
+        ByteArrayOutputStream bos = new ByteArrayOutputStream();
+        ObjectOutputStream oos = new ObjectOutputStream(bos);
+        oos.writeUTF(sessionId);
+        oos.flush();
+        oos.close();
+        return bos.toByteArray();
+    }
+
+    /**
+     * Load sessionID
+     * @throws IOException if an input/output error occurs
+     */
+    protected String deserializeSessionId(byte[] data) throws IOException {
+        ReplicationStream ois = getReplicationStream(data);
+        String sessionId = ois.readUTF();
+        ois.close();
+        return sessionId;
+    }
+
     /**
      * Load Deltarequest from external node
      * Load the Class at container classloader
@@ -689,6 +774,9 @@ public class DeltaManager extends ClusterManagerBase{
                     if (log.isWarnEnabled()) log.warn(sm.getString("deltaManager.loading.existing.session",session.getIdInternal()));
                 }
                 add(session);
+                if (notifySessionListenersOnReplication) {
+                    session.tellNew();
+                }
             }
         } catch (ClassNotFoundException e) {
             log.error(sm.getString("deltaManager.loading.cnfe", e), e);
@@ -1075,7 +1163,8 @@ public class DeltaManager extends ClusterManagerBase{
                 case SessionMessage.EVT_SESSION_CREATED: 
                 case SessionMessage.EVT_SESSION_EXPIRED: 
                 case SessionMessage.EVT_SESSION_ACCESSED:
-                case SessionMessage.EVT_SESSION_DELTA: {
+                case SessionMessage.EVT_SESSION_DELTA:
+                case SessionMessage.EVT_CHANGE_SESSION_ID: {
                     synchronized(receivedMessageQueue) {
                         if(receiverQueue) {
                             receivedMessageQueue.add(msg);
@@ -1106,9 +1195,33 @@ public class DeltaManager extends ClusterManagerBase{
      * @return a SessionMessage to be sent,
      */
     public ClusterMessage requestCompleted(String sessionId) {
+        return requestCompleted(sessionId, false);
+    }
+
+    /**
+     * When the request has been completed, the replication valve will notify
+     * the manager, and the manager will decide whether any replication is
+     * needed or not. If there is a need for replication, the manager will
+     * create a session message and that will be replicated. The cluster
+     * determines where it gets sent.
+     * 
+     * Session expiration also calls this method, but with expires == true.
+     * 
+     * @param sessionId -
+     *            the sessionId that just completed.
+     * @param expires -
+     *            whether this method has been called during session expiration
+     * @return a SessionMessage to be sent,
+     */
+    public ClusterMessage requestCompleted(String sessionId, boolean expires) {
         DeltaSession session = null;
         try {
             session = (DeltaSession) findSession(sessionId);
+            if (session == null) {
+                // A parallel request has called session.invalidate() which has
+                // remove the session from the Manager.
+                return null;
+            }
             DeltaRequest deltaRequest = session.getDeltaRequest();
             session.lock();
             SessionMessage msg = null;
@@ -1127,7 +1240,7 @@ public class DeltaManager extends ClusterManagerBase{
                 }  
             }
             if(!isDeltaRequest) {
-                if(!session.isPrimarySession()) {               
+                if(!expires && !session.isPrimarySession()) {
                     counterSend_EVT_SESSION_ACCESSED++;
                     msg = new SessionMessageImpl(getName(),
                                                  SessionMessage.EVT_SESSION_ACCESSED, 
@@ -1143,11 +1256,13 @@ public class DeltaManager extends ClusterManagerBase{
                     log.debug(sm.getString("deltaManager.createMessage.delta",getName(), sessionId));
                 }
             }
-            session.setPrimarySession(true);
+            if (!expires)
+                session.setPrimarySession(true);
             //check to see if we need to send out an access message
-            if ((msg == null)) {
+            if (!expires && (msg == null)) {
                 long replDelta = System.currentTimeMillis() - session.getLastTimeReplicated();
-                if (replDelta > (getMaxInactiveInterval() * 1000)) {
+                if (session.getMaxInactiveInterval() >=0 && 
+                        replDelta > (session.getMaxInactiveInterval() * 1000)) {
                     counterSend_EVT_SESSION_ACCESSED++;
                     msg = new SessionMessageImpl(getName(),
                                                  SessionMessage.EVT_SESSION_ACCESSED, 
@@ -1162,7 +1277,10 @@ public class DeltaManager extends ClusterManagerBase{
             }
 
             //update last replicated time
-            if (msg != null) session.setLastTimeReplicated(System.currentTimeMillis());
+            if (msg != null){
+               session.setLastTimeReplicated(System.currentTimeMillis());
+               msg.setTimestamp(session.getLastTimeReplicated());
+            }
             return msg;
         } catch (IOException x) {
             log.error(sm.getString("deltaManager.createMessage.unableCreateDeltaRequest",sessionId), x);
@@ -1178,10 +1296,24 @@ public class DeltaManager extends ClusterManagerBase{
     public synchronized void resetStatistics() {
         processingTime = 0 ;
         expiredSessions = 0 ;
+        synchronized (sessionCreationTiming) {
+            sessionCreationTiming.clear();
+            while (sessionCreationTiming.size() <
+                    ManagerBase.TIMING_STATS_CACHE_SIZE) {
+                sessionCreationTiming.add(null);
+            }
+        }
+        synchronized (sessionExpirationTiming) {
+            sessionExpirationTiming.clear();
+            while (sessionExpirationTiming.size() <
+                    ManagerBase.TIMING_STATS_CACHE_SIZE) {
+                sessionExpirationTiming.add(null);
+            }
+        }
         rejectedSessions = 0 ;
         sessionReplaceCounter = 0 ;
         counterNoStateTransfered = 0 ;
-        maxActive = getActiveSessions() ;
+        setMaxActive(getActiveSessions());
         sessionCounter = getActiveSessions() ;
         counterReceive_EVT_ALL_SESSION_DATA = 0;
         counterReceive_EVT_GET_ALL_SESSIONS = 0;
@@ -1190,6 +1322,7 @@ public class DeltaManager extends ClusterManagerBase{
         counterReceive_EVT_SESSION_DELTA = 0 ;
         counterReceive_EVT_SESSION_EXPIRED = 0 ;
         counterReceive_EVT_ALL_SESSION_TRANSFERCOMPLETE = 0;
+        counterReceive_EVT_CHANGE_SESSION_ID = 0;
         counterSend_EVT_ALL_SESSION_DATA = 0;
         counterSend_EVT_GET_ALL_SESSIONS = 0;
         counterSend_EVT_SESSION_ACCESSED = 0 ;
@@ -1197,6 +1330,7 @@ public class DeltaManager extends ClusterManagerBase{
         counterSend_EVT_SESSION_DELTA = 0 ;
         counterSend_EVT_SESSION_EXPIRED = 0 ;
         counterSend_EVT_ALL_SESSION_TRANSFERCOMPLETE = 0;
+        counterSend_EVT_CHANGE_SESSION_ID = 0;
         
     }
    
@@ -1221,6 +1355,7 @@ public class DeltaManager extends ClusterManagerBase{
     protected void sessionExpired(String id) {
         counterSend_EVT_SESSION_EXPIRED++ ;
         SessionMessage msg = new SessionMessageImpl(getName(),SessionMessage.EVT_SESSION_EXPIRED, null, id, id+ "-EXPIRED-MSG");
+        msg.setTimestamp(System.currentTimeMillis());
         if (log.isDebugEnabled()) log.debug(sm.getString("deltaManager.createMessage.expire",getName(), id));
         send(msg);
     }
@@ -1334,6 +1469,10 @@ public class DeltaManager extends ClusterManagerBase{
                    handleSESSION_DELTA(msg,sender);
                    break;
                 }
+                case SessionMessage.EVT_CHANGE_SESSION_ID: {
+                    handleCHANGE_SESSION_ID(msg,sender);
+                    break;
+                 }
                 default: {
                     //we didn't recognize the message type, do nothing
                     break;
@@ -1431,12 +1570,9 @@ public class DeltaManager extends ClusterManagerBase{
         session.setPrimarySession(false);
         session.setCreationTime(msg.getTimestamp());
         // use container maxInactiveInterval so that session will expire correctly in case of primary transfer
-        session.setMaxInactiveInterval(getMaxInactiveInterval());
+        session.setMaxInactiveInterval(getMaxInactiveInterval(), false);
         session.access();
-        if(notifySessionListenersOnReplication)
-            session.setId(msg.getSessionID());
-        else
-            session.setIdInternal(msg.getSessionID());
+        session.setId(msg.getSessionID(), notifySessionListenersOnReplication);
         session.resetDeltaRequest();
         session.endAccess();
 
@@ -1479,10 +1615,9 @@ public class DeltaManager extends ClusterManagerBase{
             sendSessions(sender, currentSessions, findSessionTimestamp);
         } else {
             // send session at blocks
-            int len = currentSessions.length < getSendAllSessionsSize() ? currentSessions.length : getSendAllSessionsSize();
-            Session[] sendSessions = new Session[len];
             for (int i = 0; i < currentSessions.length; i += getSendAllSessionsSize()) {
-                len = i + getSendAllSessionsSize() > currentSessions.length ? currentSessions.length - i : getSendAllSessionsSize();
+                int len = i + getSendAllSessionsSize() > currentSessions.length ? currentSessions.length - i : getSendAllSessionsSize();
+                Session[] sendSessions = new Session[len];
                 System.arraycopy(currentSessions, i, sendSessions, 0, len);
                 sendSessions(sender, sendSessions,findSessionTimestamp);
                 if (getSendAllSessionsWaitTime() > 0) {
@@ -1501,6 +1636,29 @@ public class DeltaManager extends ClusterManagerBase{
         cluster.send(newmsg, sender);
     }
 
+    /**
+     * handle receive change sessionID at other node
+     * @param msg
+     * @param sender
+     * @throws IOException
+     */
+    protected void handleCHANGE_SESSION_ID(SessionMessage msg,Member sender) throws IOException {
+        counterReceive_EVT_CHANGE_SESSION_ID++;
+        DeltaSession session = (DeltaSession) findSession(msg.getSessionID());
+        if (session != null) {
+            String newSessionID = deserializeSessionId(msg.getSession());
+            session.setPrimarySession(false);
+            session.setId(newSessionID, false);
+            if (notifyContainerListenersOnReplication) {
+                Container c = getContainer();
+                if (c instanceof StandardContext) {
+                    ((StandardContext) getContainer()).fireContainerEvent(
+                            Context.CHANGE_SESSION_ID_EVENT,
+                            new String[] {msg.getSessionID(), newSessionID});
+                }
+            }
+        }
+    }
 
     /**
      * send a block of session to sender
@@ -1528,6 +1686,7 @@ public class DeltaManager extends ClusterManagerBase{
         result.expireSessionsOnShutdown = expireSessionsOnShutdown;
         result.notifyListenersOnReplication = notifyListenersOnReplication;
         result.notifySessionListenersOnReplication = notifySessionListenersOnReplication;
+        result.notifyContainerListenersOnReplication = notifyContainerListenersOnReplication;
         result.stateTransferTimeout = stateTransferTimeout;
         result.sendAllSessions = sendAllSessions;
         result.sendClusterDomainOnly = sendClusterDomainOnly ;
