@@ -39,49 +39,62 @@ import org.apache.catalina.ha.session.DeltaSession;
 import org.apache.catalina.connector.Request;
 import org.apache.catalina.connector.Response;
 import org.apache.catalina.session.ManagerBase;
+import org.apache.catalina.session.PersistentManager;
 import org.apache.catalina.util.LifecycleSupport;
 import org.apache.catalina.util.StringManager;
 import org.apache.catalina.valves.ValveBase;
 
 /**
  * Valve to handle Tomcat jvmRoute takeover using mod_jk module after node
- * failure. After a node crashed the next request going to other cluster node.
- * Now the answering from apache is slower ( make some error handshaking. Very
- * bad with apache at my windows.). We rewrite now the cookie jsessionid
- * information to the backup cluster node. After the next response all client
- * request goes direct to the backup node. The change sessionid send also to all
- * other cluster nodes. Well, now the session stickyness work directly to the
- * backup node and traffic don't go back too restarted cluster nodes!
+ * failure. After a node crashes, subsequent requests go to other cluster nodes.
+ * That incurs a drop in performance. When this Valve is enabled on a backup
+ * node and sees a request, which was intended for another (thus failed) node,
+ * it will rewrite the cookie jsessionid information to use the route to this
+ * backup cluster node, that answered the request. After the response is
+ * delivered to the client, all subsequent client requests will go directly to
+ * the backup node. The change of sessionid is also sent to all other cluster
+ * nodes. After all that, the session stickyness will work directly to the
+ * backup node and the traffic will not go back to the failed node after it is
+ * restarted!
  * 
- * At all cluster node you must configure the as ClusterListener since 5.5.10
- * {@link org.apache.catalina.ha.session.JvmRouteSessionIDBinderListener JvmRouteSessionIDBinderListener}
- * or before with
- * org.apache.catalina.ha.session.JvmRouteSessionIDBinderListenerLifecycle.
+ * <p>
+ * For this valve to function correctly, so that all nodes of the cluster
+ * receive the sessionid change notifications that it generates, the following
+ * ClusterListener MUST be configured at all nodes of the cluster:
+ * {@link org.apache.catalina.ha.session.JvmRouteSessionIDBinderListener
+ * JvmRouteSessionIDBinderListener} since Tomcat 5.5.10, and both
+ * JvmRouteSessionIDBinderListener and JvmRouteSessionIDBinderLifecycleListener
+ * for earlier versions of Tomcat.
  * 
+ * <p>
  * Add this Valve to your host definition at conf/server.xml .
  * 
  * Since 5.5.10 as direct cluster valve:<br/>
+ * 
  * <pre>
  *  &lt;Cluster&gt;
  *  &lt;Valve className=&quot;org.apache.catalina.ha.session.JvmRouteBinderValve&quot; /&gt;  
  *  &lt;/Cluster&gt;
  * </pre>
+ * 
  * <br />
  * Before 5.5.10 as Host element:<br/>
+ * 
  * <pre>
- *  &lt;Hostr&gt;
+ *  &lt;Host&gt;
  *  &lt;Valve className=&quot;org.apache.catalina.ha.session.JvmRouteBinderValve&quot; /&gt;  
- *  &lt;/Hostr&gt;
+ *  &lt;/Host&gt;
  * </pre>
  * 
- * Trick:<br/>
- * You can enable this mod_jk turnover mode via JMX before you drop a node to all backup nodes!
- * Set enable true on all JvmRouteBinderValve backups, disable worker at mod_jk 
- * and then drop node and restart it! Then enable mod_jk Worker and disable JvmRouteBinderValves again. 
- * This use case means that only requested session are migrated.
+ * <em>A Trick:</em><br/>
+ * You can enable this mod_jk turnover mode via JMX before you drop a node to
+ * all backup nodes! Set enable true on all JvmRouteBinderValve backups, disable
+ * worker at mod_jk and then drop node and restart it! Then enable mod_jk worker
+ * and disable JvmRouteBinderValves again. This use case means that only
+ * requested sessions are migrated.
  * 
  * @author Peter Rossbach
- * @version $Revision: 750895 $ $Date: 2009-03-06 14:51:53 +0100 (Fri, 06 Mar 2009) $
+ * @version $Id: JvmRouteBinderValve.java 1135497 2011-06-14 11:41:45Z markt $
  */
 public class JvmRouteBinderValve extends ValveBase implements ClusterValve, Lifecycle {
 
@@ -196,14 +209,18 @@ public class JvmRouteBinderValve extends ValveBase implements ClusterValve, Life
             ServletException {
 
          if (getEnabled() 
-             && getCluster() != null
              && request.getContext() != null
              && request.getContext().getDistributable() ) {
              // valve cluster can access manager - other cluster handle turnover 
              // at host level - hopefully!
              Manager manager = request.getContext().getManager();
-             if (manager != null && manager instanceof ClusterManager
-                     && getCluster().getManager(((ClusterManager)manager).getName()) != null)
+
+             if (manager != null && (
+                     (manager instanceof ClusterManager
+                       && getCluster() != null
+                       && getCluster().getManager(((ClusterManager)manager).getName()) != null)
+                     ||
+                     (manager instanceof PersistentManager)))
                  handlePossibleTurnover(request, response);
         }
         // Pass this request on to the next valve in our pipeline
@@ -354,16 +371,19 @@ public class JvmRouteBinderValve extends ValveBase implements ClusterValve, Life
             Response response, String sessionId, String newSessionID, Session catalinaSession) {
         lifecycle.fireLifecycleEvent("Before session migration",
                 catalinaSession);
-        // FIXME: setId trigger session Listener, but only chance to registiert manager with correct id!
-        catalinaSession.setId(newSessionID);
+        catalinaSession.setId(newSessionID, false);
         // FIXME: Why we remove change data from other running request?
         // setId also trigger resetDeltaRequest!!
         if (catalinaSession instanceof DeltaSession)
             ((DeltaSession) catalinaSession).resetDeltaRequest();
         changeRequestSessionID(request, response, sessionId, newSessionID);
-        // now sending the change to all other clusternode!
-        ClusterManager manager = (ClusterManager)catalinaSession.getManager();
-        sendSessionIDClusterBackup(manager,request,sessionId, newSessionID);
+
+        if (getCluster() != null) {
+            // now sending the change to all other clusternode!
+            ClusterManager manager = (ClusterManager)catalinaSession.getManager();
+            sendSessionIDClusterBackup(manager,request,sessionId, newSessionID);
+        }
+
         lifecycle.fireLifecycleEvent("After session migration", catalinaSession);
         if (log.isDebugEnabled()) {
             log.debug(sm.getString("jvmRoute.changeSession", sessionId,
@@ -381,9 +401,8 @@ public class JvmRouteBinderValve extends ValveBase implements ClusterValve, Life
      *            new session id for node migration
      */
     protected void changeRequestSessionID(Request request, Response response, String sessionId, String newSessionID) {
-        request.setRequestedSessionId(newSessionID);
-        if(request.isRequestedSessionIdFromCookie())
-            setNewSessionCookie(request, response,newSessionID);
+        request.changeSessionId(newSessionID);
+
         // set orginal sessionid at request, to allow application detect the
         // change
         if (sessionIdAttribute != null && !"".equals(sessionIdAttribute)) {
@@ -426,6 +445,8 @@ public class JvmRouteBinderValve extends ValveBase implements ClusterValve, Life
      * @param request current request
      * @param response Tomcat Response
      * @param sessionId The session id
+     * 
+     * @deprecated Use {@link Request#changeSessionId(String)}
      */
     protected void setNewSessionCookie(Request request,
                                        Response response, String sessionId) {
@@ -433,29 +454,46 @@ public class JvmRouteBinderValve extends ValveBase implements ClusterValve, Life
             Context context = request.getContext();
             if (context.getCookies()) {
                 // set a new session cookie
-                Cookie newCookie = new Cookie(Globals.SESSION_COOKIE_NAME,
-                        sessionId);
+                String scName = context.getSessionCookieName();
+                if (scName == null) {
+                    scName = Globals.SESSION_COOKIE_NAME;
+                }
+                Cookie newCookie = new Cookie(scName, sessionId);
+                
                 newCookie.setMaxAge(-1);
+                
                 String contextPath = null;
-                if (!response.getConnector().getEmptySessionPath()
-                        && (context != null)) {
-                    contextPath = context.getEncodedPath();
+                if (!response.getConnector().getEmptySessionPath() &&
+                        (context != null)) {
+                    if (context.getSessionCookiePath() != null) {
+                        contextPath = context.getSessionCookiePath();
+                    } else {
+                        contextPath = context.getEncodedPath();
+                    }
                 }
                 if ((contextPath != null) && (contextPath.length() > 0)) {
                     newCookie.setPath(contextPath);
                 } else {
                     newCookie.setPath("/");
                 }
+                
+                if (context.getSessionCookieDomain() != null) {
+                    newCookie.setDomain(context.getSessionCookieDomain());
+                }
+
                 if (request.isSecure()) {
                     newCookie.setSecure(true);
                 }
+
                 if (log.isDebugEnabled()) {
-                    log.debug(sm.getString("jvmRoute.newSessionCookie",
-                            sessionId, Globals.SESSION_COOKIE_NAME, newCookie
-                                    .getPath(), new Boolean(newCookie
-                                    .getSecure())));
+                    Object[] args = new Object[] {sessionId,
+                            newCookie.getName(),
+                            newCookie.getPath(),
+                            new Boolean(newCookie.getSecure()),
+                            new Boolean(context.getUseHttpOnly())};
+                    log.debug(sm.getString("jvmRoute.newSessionCookie", args));
                 }
-                response.addCookie(newCookie);
+                response.addCookieInternal(newCookie, context.getUseHttpOnly());
             }
         }
     }
@@ -533,13 +571,12 @@ public class JvmRouteBinderValve extends ValveBase implements ClusterValve, Life
                 }
             }
         }
-        if (cluster == null) {
-            throw new RuntimeException("No clustering support at container "
-                    + container.getName());
-        }
         
-        if (log.isInfoEnabled())
+        if (log.isInfoEnabled()) {
             log.info(sm.getString("jvmRoute.valve.started"));
+            if (cluster == null)
+                log.info(sm.getString("jvmRoute.noCluster"));
+        }
 
     }
 
